"""
# Description

This module provides straightforward functions to plot QRotor data.


# Index

`reduced_energies()`  
`energies_DEV()`      NOT IMPLEMENTED  
`energy_DEV()`        NOT IMPLEMENTED  
`convergence_DEV()`   NOT IMPLEMENTED  
`eigenvectors_DEV()`  NOT IMPLEMENTED  

---
"""


from .classes import *
import matplotlib.pyplot as plt


def reduced_energies(data:QExp):
    """Plots the reduced energy of the system, E/B, vs the reduced potential energy, V/B"""
    number_of_levels = data.systems[0].E_levels
    x = []
    for system in data.systems:
        x.append(system.potential_max_B)
    for i in range(number_of_levels):
        y = []
        for system in data.systems:
            y.append(system.eigenvalues_B[i])
        plt.plot(x, y, marker='', linestyle='-')
    plt.xlabel('V$_{B}$ / B')
    plt.ylabel('E / B')
    plt.title(data.comment)
    plt.show()


def energies_DEV(data:QExp):
    '''Plots the energy in separated plots. NOT YET IMPLEMENTED'''
    if data.separate_plots:
        for variables, solutions in zip(data.variables, data.solutions):
            new_data = QExp()
            new_data.comment = variables.comment
            new_data.variables.append(variables)
            new_data.solutions.append(solutions)
            energy(new_data)
    else:
        # Group data with the same potential_values and different element
        grouped_data = data.group_by_potential_and_atoms()
        for i, new_data in enumerate(grouped_data):
            energy(new_data)


def energy_DEV(data:QExp):
    '''Plots the energy of the system. NOT YET IMPLEMENTED'''
    V_colors = ['C0'] #...
    E_colors = ['red', 'purple', 'grey']  # To extend...
    E_linestyles = ['--', ':', '-.']
    edgecolors = ['tomato', 'purple', 'grey']

    V_linestyle = '-'

    units = data.units
    E_units = 'meV'
    if 'meV' in units or 'mev' in units:
        E_units = 'meV'
    elif 'eV' in units or 'ev' in units:
        E_units = 'eV'

    title = data.comment
    ylabel_text = f'Energy / {E_units}'
    xlabel_text = 'Angle / radians'

    plt.figure(figsize=(10, 6))
    plt.xlabel(xlabel_text)
    plt.ylabel(ylabel_text)

    plt.title(title)
    if not data.comment or (len(data.variables) == 1 and data.variables[0].comment):
        plt.title(f'{data.variables[0].comment}')

    plt.xticks([0, np.pi/2, np.pi, 3*np.pi/2, 2*np.pi],
                ['0', r'$\frac{\pi}{2}$', r'$\pi$', r'$\frac{3\pi}{2}$', r'$2\pi$'])

    unique_potentials = []
    unique_elements = []
    for i, (variables, solutions) in enumerate(zip(data.variables, data.solutions)):
        V_color = V_colors[i % len(V_colors)]
        E_color = E_colors[i % len(E_colors)]
        E_linestyle = E_linestyles[i % len(E_linestyles)]
        edgecolor = edgecolors[i % len(edgecolors)]

        # Plot potential energy if it is unique
        if not any(np.array_equal(variables.potential_values, element) for element in unique_potentials):
            unique_potentials.append(variables.potential_values)
            plt.plot(variables.grid, variables.potential_values, color=V_color, linestyle=V_linestyle)

        # Plot eigenvalues
        if solutions.eigenvalues is not None:
            text_offset = 3 * len(unique_elements)
            if variables.element not in unique_elements:
                unique_elements.append(variables.element)
            for j, energy in enumerate(solutions.eigenvalues):
                plt.axhline(y=energy, color=E_color, linestyle=E_linestyle)
                plt.text(j%3*0.9 + text_offset, energy, f'$E_{{{j}}}$ = {round(energy,4):.04f}', va='top', bbox=dict(edgecolor=edgecolor, boxstyle='round,pad=0.2', facecolor='white', alpha=0.8))
            if len(data.elements()) > 1:
                plt.plot([], [], color=E_color, label=f'{variables.element} Energies')  # Add to legend

    if len(data.elements()) > 1:
        plt.subplots_adjust(right=0.85)
        plt.legend(bbox_to_anchor=(1.1, 0.5), loc='center', fontsize='small')

    plt.show()


def convergence_DEV(data:QExp):
    '''Plots the energy convergence of the system. NOT YET IMPLEMENTED'''
    fig, ax1 = plt.subplots(figsize=(10, 6))

    E_color = 'C0'
    runtime_color = 'C3'
    yaxes_color = E_color

    converged_color_line = 'lightgrey'

    units = data.variables[0].units
    E_units = 'meV'
    if 'meV' in units or 'mev' in units:
        E_units = 'meV'
    elif 'eV' in units or 'ev' in units:
        E_units = 'eV'

    title = data.comment
    ylabel_text = f'Energy / {E_units}'
    xlabel_text = 'Grid Size'
    runtime_text = 'Runtime / s'

    legend_title = data.legend_title
    legend_title_position = data.legend_title_position
    check_E_threshold = data.check_E_threshold
    check_E_diff = data.check_E_diff
    check_E_level = data.check_E_level
    ideal_E = data.ideal_E
    if check_E_level is None:
        data.check_E_level = len(data.solutions[0].eigenvalues) - 1
        check_E_level = data.check_E_level
        ideal_E = data.get_ideal_E()


    textbox = dict(boxstyle='round', facecolor='white', edgecolor='lightgrey', alpha=0.5)
    textstr = ''

    textstr_position_x = 0.88
    textstr_position_y = 0.15
    textstr_alignment_h = 'right'
    textstr_alignment_v = 'bottom'

    if legend_title_position and isinstance(legend_title_position, list):
        textstr_position_x = data.legend_title_position[0]
        textstr_position_y = data.legend_title_position[1]
        textstr_alignment_h = data.legend_title_position[2]
        textstr_alignment_v = data.legend_title_position[3]

    energies = data.energies()
    energies_transposed = np.transpose(energies)
    plotted_energies = energies_transposed[check_E_level]
    gridsizes = data.gridsizes()
    runtimes = data.runtimes()

    if check_E_diff:
        plotted_energies = np.abs(plotted_energies - ideal_E)
        ylabel_text = 'Energy offset / |meV|'
        textstr_position_x = 0.5
        textstr_position_y = 0.85
        textstr_alignment_v = 'top'
        textstr_alignment_h = 'center'
    
    if not any(runtimes):
        yaxes_color = 'black'

    ax1.plot(gridsizes, plotted_energies, marker='o', linestyle='-', color=E_color)
    ax1.set_xlabel(xlabel_text)
    ax1.set_ylabel(ylabel_text, color=yaxes_color)
    ax1.tick_params(axis='y', labelcolor=yaxes_color)

    if ideal_E is not None:
        if check_E_diff:
            ax1.axhline(y=0, color='grey', linestyle='--')
        else:
            ax1.axhline(y=ideal_E, color='grey', linestyle='--')
            textstr += f'Ideal  E={ideal_E:.4f}\n'
    
    if check_E_threshold and (ideal_E is not None):
        if check_E_diff:
            abs_energies = energy
        else:
            abs_energies = np.abs(plotted_energies - ideal_E)
        for i, energy in enumerate(abs_energies):
            if energy < check_E_threshold:
                #ax1.plot(gridsizes[i], plotted_energies[i], marker=converged_marker, color=E_converged_color)
                ax1.axvline(x=gridsizes[i], color=converged_color_line, linestyle='--')
                textstr += f'Convergence threshold:  {check_E_threshold}\n'
                lower_limit, _ = ax1.get_ylim()
                ax1.text(gridsizes[i], lower_limit, str(gridsizes[i]), fontsize=10, verticalalignment='bottom', horizontalalignment='center')
                break

    if any(runtimes):
        ax2 = ax1.twinx()  # instantiate a second y-axis that shares the same x-axis
        ax2.set_ylabel(runtime_text, color=runtime_color)  # we already handled the x-label with ax1
        ax2.plot(gridsizes, runtimes, marker='o', linestyle='-', color=runtime_color)
        ax2.tick_params(axis='y', labelcolor=runtime_color)
        for i, energy in enumerate(plotted_energies):
            textstr += f'N={gridsizes[i]}   E={round(energy,4):.04f}   t={round(runtimes[i],2):.02f}'
            if i < len(plotted_energies) - 1:
                textstr += '\n'

    else:
        for i, energy in enumerate(plotted_energies):
            textstr += f'N={gridsizes[i]}   E={round(energy,4):.04f}\n'

    if legend_title is not False:
        if isinstance(legend_title, str):
            textstr = legend_title + '\n' + textstr
        fig.text(textstr_position_x, textstr_position_y, textstr, fontsize=10, verticalalignment=textstr_alignment_v, horizontalalignment=textstr_alignment_h, bbox=textbox)

    plt.title(title)
    plt.show()









def eigenvectors_DEV(data:QExp, levels=None, squared=False, scaling_factor=1):

    xlabel = 'Angle / radians'
    ylabel = 'Energy / meV'
    title = data.title
    V_color = 'lightblue'
    V_label = 'Potential'

    #energy_color = 'red'
    energy_edgecolor = 'lightgrey'
    energy_linestyle = ':'
    energy_label = 'E'

    eigenvector_linestyle = '--'

    # To square the eigenvectors
    if squared:
        eigenvector_label = 'Eigenvect$^2$ '
        square = 2
    else:
        eigenvector_label = 'Eigenvect '
        square = 1
    
    for i, potential in enumerate(data.set_of_potentials):

        # Transpose the 2D array so that each inner array represents a different eigenvector
        eigenvectors_transposed = np.transpose(data.set_of_eigenvectors[i])

        # Plot potential energy
        plt.figure(figsize=(10, 6))
        plt.plot(data.x, potential, color=V_color, label=V_label)
        plt.xlabel(xlabel)
        plt.ylabel(ylabel)
        plt.title(f'{title} (#' + str(i+1) + ')' )
        if len(data.set_of_potentials) == 1:
            plt.title(f'{title}')
        plt.xticks([0, np.pi/2, np.pi, 3*np.pi/2, 2*np.pi],
                   ['0', r'$\frac{\pi}{2}$', r'$\pi$', r'$\frac{3\pi}{2}$', r'$2\pi$'])
        for j, energy in enumerate(data.set_of_energies[i]):
            if levels is not None and j not in levels:
                continue

            color = 'C' + str(j)

            E_label = energy_label + str(j)
            plt.axhline(y=energy, linestyle=energy_linestyle, color=color, label=E_label)
            plt.text(j%3*0.9, energy, f'E$_{j}$ = {energy:.4f}', va='top', bbox=dict(edgecolor=energy_edgecolor, boxstyle='round,pad=0.2', facecolor='white', alpha=0.8))

            eigenvect_label = eigenvector_label + str(j)
            eigenvector = scaling_factor*eigenvectors_transposed[j]**square
            plt.plot(data.x, eigenvector, linestyle=eigenvector_linestyle, label=eigenvect_label, color=color)

        plt.subplots_adjust(right=0.85)
        plt.legend(bbox_to_anchor=(1.1, 0.5), loc='center', fontsize='small')
        plt.text(1.03, 0.9, f'Eigenvects\nscaled x{scaling_factor}', transform=plt.gca().transAxes)
        plt.show()






"""
# ⚠️ UNDER DEVELOPMENT

THIS PACKAGE IS STILL UNDER HEAVY DEVELOPMENT, DON'T USE IT

# QRotor
 
The QRotor subpackage is used to study the energy levels of quantum rotations, such as methyl and amine groups.

# Index

| | |
| --- | --- |
| `aton.qrotor.classes`   | Definition of the `QSys` and `QExp` classes |
| `aton.qrotor.constants` | Bond lengths and inertias |
| `aton.qrotor.rotate`    | Rotate specific atoms from structural files |
| `aton.qrotor.potential` | Potential definitions and loading functions |
| `aton.qrotor.solve`     | Solve rotation eigenvalues and eigenvectors |
| `aton.qrotor.plot`      | Plotting functions |

"""


from .classes import QSys, QExp
from .constants import *
from . import rotate
from . import potential
from . import solve
from . import plot        ###### TODO: update




"""
.. include:: ../_README_temp.md
"""


from ._version import __version__ as version
from . import st
from . import phys
from . import txt
from . import interface
from . import spx
from . import qrotor




"""
# Description

This module contains functions to calculate the actual `potential_values` of the system.


# Index

| | |
| --- | --- |
| `load()`        | Load a system with a custom potential from a potential file |
| `from_qe()`     | Creates a potential data file from Quantum ESPRESSO outputs |
| `interpolate()` | Interpolates the current `QSys.potential_values` to a new `QSys.gridsize` |
| `solve()`       | Solve the potential values based on the potential name |
| `zero()`        | Zero potential |
| `sine()`        | Sine potential |
| `titov2023()`   | Potential of the hidered methyl rotor, as in titov2023. |

---
"""


from .classes import *
from . import constants
import numpy as np
import os
from scipy.interpolate import CubicSpline
import aton.st.alias as alias
import aton.st.file as file
import aton.interface.qe as qe
import aton.phys as phys


def load(
        filepath,
        system:QSys=None,
        angle:str='deg',
        energy:str='eV',
        ) -> QSys:
    """Read a potential rotational energy dataset.

    The file should contain two columns with `angle` and `potential` values.
    Degrees and eV are assumed as default units unless stated in `angle` and `energy`.
    """
    file_path = file.get(filepath)
    system = QSys() if system is None else system
    with open(file_path, 'r') as f:
        lines = f.readlines()
    positions = []
    potentials = []
    for line in lines:
        if line.startswith('#'):
            continue
        position, potential = line.split()
        positions.append(float(position))
        potentials.append(float(potential))
    # Save angles to numpy arrays
    if angle.lower() in alias.units['deg']:
        positions = np.radians(positions)
    elif angle.lower() in alias.units['rad']:
        positions = np.array(positions)
    else:
        raise ValueError(f"Angle unit '{angle}' not recognized.")
    # Save energies to numpy arrays
    if energy.lower() in alias.units['meV']:
        potentials = np.array(potentials) * 1000
    elif energy.lower() in alias.units['ev']:
        potentials = np.array(potentials)
    else:
        raise ValueError(f"Energy unit '{energy}' not recognized.")
    # Set the system
    system.grid = np.array(positions)
    system.gridsize = len(positions)
    system.potential_values = np.array(potentials)
    return system


def from_qe(
        folder=None,
        filters:str=None,
        output:str='potential.dat',
        ) -> None:
    """Creates a potential data file from Quantum ESPRESSO outputs.

    The angle in degrees is extracted from the output filenames,
    which must follow `whatever_ANGLE.out`.

    Outputs from SCF calculations must be located in the provided `folder` (CWD if None),
    and can be filtered with `filters`.
    The `output` name is `potential.dat` by default.
    """
    folder = file.get_dir(folder)
    files = file.get_list(folder=folder, filters=filters, abspath=True)
    potential_data = '# Angle / deg, Potential / eV\n'
    for filepath in files:
        filepath = file.get(filepath=filepath, filters='.out', return_anyway=True)
        if not filepath:  # Not an output file, skip it
            continue
        content = qe.read_out(filepath)
        if not content['Success']:  # Ignore unsuccessful calculations
            continue
        energy = content['Energy'] * phys.Ry_to_eV
        filename = os.path.basename(filepath)
        splits = filename.split('_')
        angle = splits[-1]
        potential_data += f'{angle}, {energy}\n'
    with open(output, 'w') as f:
        f.write(potential_data)
    return None


def interpolate(system:QSys) -> QSys:
    """Interpolates the current `aton.qrotor.classes.QSys.potential_values` to a new grid of size `aton.qrotor.classes.QSys.gridsize`."""
    V = system.potential_values
    grid = system.grid
    gridsize = system.gridsize
    new_grid = np.linspace(0, 2*np.pi, gridsize)
    cubic_spline = CubicSpline(grid, V)
    new_V = cubic_spline(new_grid)
    system.grid = new_grid
    system.potential_values = new_V
    return system


# Redirect to the desired potential energy function
def solve(system:QSys):
    """Solves `aton.qrotor.classes.QSys.potential_values`, according to the `aton.qrotor.classes.QSys.potential_name`."""
    if system.potential_name.lower() == 'titov2023':
        return titov2023(system)
    elif system.potential_name.lower() == 'zero':
        return zero(system)
    elif system.potential_name.lower() == 'sine':
        return sine(system)
    elif system.potential_name.lower() == 'test':
        return test(system)
    elif system.potential_values:  # Re
        return system.potential_values
    else:
        raise ValueError(f'Unrecognised potential_name ({system.potential_name}) and no potential_values found in system!')


def zero(system:QSys):
    """Zero potential."""
    x = system.grid
    return 0 * x


def sine(system:QSys):
    """Sine potential.

    If potential_constants are provided, returns:
    $C_0 + C_1 sin(3x + C_2)$.
    """
    x = system.grid
    C = system.potential_constants
    C0 = 0
    C1 = 1
    C2 = 0
    if C:
        if len(C) > 0:
            C0 = C[0]
        if len(C) > 1:
            C1 = C[1]
        if len(C) > 2:
            C2 = C[2]
    return C0 + C1 * np.sin(3*x + C2)


def titov2023(system:QSys):
    """Potential energy function of the hindered methyl rotor, from titov2023."""
    x = system.grid
    C = system.potential_constants
    if C is None:
        C = constants.constants_titov2023[0]
    return C[0] + C[1] * np.sin(3*x) + C[2] * np.cos(3*x) + C[3] * np.sin(6*x) + C[4] * np.cos(6*x)




"""
# Description

Functions to handle Slurm calls, to run calculations in clusters.  


# Index

| | |
| --- | --- |
| `sbatch()`         | Sbatch all calculations |
| `scancel()`        | Scancel all calculations |
| `check_template()` | Checks that the slurm template is OK, and provides an example if not |

---
"""


import os
import aton.st.call as call
import aton.st.file as file
import aton.txt.find as find
import aton.txt.edit as edit
from aton._version import __version__


def sbatch(
        prefix:str='',
        template:str='template.slurm',
        in_ext:str='.in',
        out_ext:str='.out',
        folder=None,
        files:list=[],
        testing:bool=False,
    ) -> None:
    """Sbatch all the calculations at once.

    Calculation names should follow `prefix_ID.ext`,
    with `prefix` as the common name across calculations,
    followed by the calculation ID, used as JOB_NAME.
    The extensions from `in_ext` and `out_ext` ('.in' and '.out' by default)
    will be used for the INPUT_FILE and OUTPUT_FILE of the slurm template.

    The slurm template, `template.slurm` by default,
    must contain the keywords JOB_ID, INPUT_FILE and OUTPUT_FILE:
    ```
    #SBATCH --job-name=JOB_NAME
    mpirun pw.x -inp INPUT_FILE > OUTPUT_FILE
    ```

    Runs from the specified `folder`, current working directory if empty.

    If more control is required, a custom list of `files` can be specified for sbatching.

    If `testing = True` it skips the final sbatching,
    just printing the commands on the screen.
    """
    print(f'\naton.interface.slurm {__version__}\n'
          'Sbatching all calculations...\n')
    key_input = 'INPUT_FILE'
    key_output = 'OUTPUT_FILE'
    key_jobname = 'JOB_NAME'
    slurm_folder = 'slurms'
    folder = call.here(folder)
    # Get input files and abort if not found
    if not files:
        inputs_raw = file.get_list(folder=folder, filters=prefix, abspath=False)
    else:
        inputs_raw = files
    inputs = []
    for filename in inputs_raw:
        if filename.endswith(in_ext):
            inputs.append(filename)
    if len(inputs) == 0:
        raise FileNotFoundError(f"Input files were not found! Expected {prefix}ID.{in_ext}")
    # Make the folder for the sbatch'ed slurm files
    call.bash(f"mkdir {slurm_folder}", folder, True, True)
    # Get the template
    slurm_file = check_template(template, folder)
    if not slurm_file:
        print(f'Aborting... Please correct {template}\n')
        return None
    for filename in inputs:
        # Get the file ID
        basename: str = os.path.basename(filename)
        basename_out: str = basename.replace(in_ext, out_ext)
        calc_id = basename.replace(prefix, '')
        calc_id = calc_id.replace(in_ext, '')
        calc_id = calc_id.replace('_', '')
        calc_id = calc_id.replace('-', '')
        calc_id = calc_id.replace('.', '')
        # Create slurm file for this supercell
        slurm_id = prefix + calc_id + '.slurm'
        # fixing dictionary with the words to replace in the template
        fixing_dict = {
            key_jobname: calc_id,
            key_input: basename,
            key_output: basename_out
        }
        edit.from_template(slurm_file, slurm_id, fixing_dict)
        if testing:
            call.bash(f"echo {slurm_id}", folder)
        else:
            call.bash(f"sbatch {slurm_id}", folder, True, False)
        call.bash(f"mv {slurm_id} {slurm_folder}", folder, False, True)  # Do not raise error if we can't move the file
    print(f'\nDone! Temporary slurm files were moved to ./{slurm_folder}/\n')


def scancel(jobs=None, folder=None, prefix:str='slurm-', sufix:str='.out') -> None:
    """Cancel all running `jobs`.
    
    If no job is provided, all jobs detected in the current folder will be cancelled.
    The jobs will be detected from the `<prefix>JOBID<sufix>` files, `slurm-JOBID.out` by default.
    """
    if jobs == None:  # Get the list of jobs
        filenames = file.get_list(folder=folder, filters=prefix, abspath=False)
        if not filenames:
            raise FileNotFoundError(f'To scancel all calculations, {prefix}JOBID{sufix} files are needed!\nConfigure the folder, as well as the prefix and sufix if necessary.')
        jobs = []
        for filename in filenames:
            filename = filename.replace(prefix, '')
            filename = filename.replace(sufix, '')
            jobs.append(filename)
    if isinstance(jobs, str):
        jobs = [jobs]
    if not isinstance(jobs, list):
        raise ValueError(f'Input jobs must be a string or a list of strings! Yours was: {type(jobs)}')
    for job in jobs:
        call.bash(f'scancel {job}', folder)
    return None


def check_template(
        template:str='template.slurm',
        folder=None,
    ) -> str:
    """Check the slurm `template` inside `folder`.

    The current working directory is used if `folder` is not provided.
    If the file does not exist or is invalid, creates a `template_EXAMPLE.slurm` file for reference.
    """
    folder = call.here(folder)
    slurm_example = 'template_EXAMPLE.slurm'
    new_slurm_file = os.path.join(folder, slurm_example)
    # Default slurm template
    content =f'''# Automatic slurm template created with aton.interface.slurm {__version__}\n# https://github.com/pablogila/ATON
#!/bin/bash
#SBATCH --partition=general
#SBATCH --qos=regular
#SBATCH --job-name=JOB_NAME
#SBATCH --ntasks=32
#SBATCH --time=1-00:00:00
#SBATCH --mem=128G
# #SBATCH --mail-user=YOUR@EMAIL
# #SBATCH --mail-type=END

module purge
module load QuantumESPRESSO/7.3-foss-2023a

mpirun pw.x -inp INPUT_FILE > OUTPUT_FILE
'''
    # If the slurm template does not exist, create one
    slurm_file = file.get(folder, template, True)
    if not slurm_file:
        with open(new_slurm_file, 'w') as f:
            f.write(content)
        print(f'!!! WARNING:  Slurm template missing, so an example was generated automatically:\n'
              f'{slurm_example}\n'
              f'PLEASE CHECK it, UPDATE it and RENAME it to {template}\n'
              'before running aton.interface.phonopy.sbatch()\n')
        return None
    # Check that the slurm file contains the INPUT_FILE, OUTPUT_FILE and JOB_NAME keywords
    key_input = find.lines(slurm_file, 'INPUT_FILE')
    key_output = find.lines(slurm_file, 'OUTPUT_FILE')
    key_jobname = find.lines(slurm_file, 'JOB_NAME')
    missing = []
    if not key_input:
        missing.append('INPUT_FILE')
    if not key_output:
        missing.append('OUTPUT_FILE')
    if not key_jobname:
        missing.append('JOB_NAME')
    if len(missing) > 0:
        with open(new_slurm_file, 'w') as f:
            f.write(content)
        print('!!! WARNING:  Some keywords were missing from your slurm template,\n'
              f'PLEASE CHECK the example at {slurm_example}\n'
              'before running aton.interface.slurm.sbatch()\n'
              f'The following keywords were missing from your {template}:')
        for key in missing:
            print(key)
        print('')
        return None
    return slurm_file  # Ready to use!




"""
# Description

Functions to move files around.


# Index

| | |
| --- | --- |
| `get()`               | Check that a file exists, and return the full path |
| `get_list()`          | Get a list of the files inside a folder, applying optional filters |
| `get_dir()`           | Get the full path of a folder or the cwd |
| `copy()`              | Copy file |
| `move()`              | Move file |
| `remove()`            | Remove file or folder |
| `rename_on_folder()`  | Batch rename files from a folder |
| `rename_on_folders()` | Barch rename files from subfolders |
| `copy_to_folders()`   | Copy files to individual subfolders |
| `save()`              | Save a Python object to a binary file, as `.aton` |
| `load()`              | Load a Python object from a binary file, as `.aton` |

---
"""


import os
import shutil
import pickle
import gzip


def get(
        filepath,
        filters=None,
        return_anyway:bool=False
        ) -> str:
    """Check if `filepath` exists, and returns its full path.

    Raises an error if the file is not found,
    unless `return_anyway = True`, in which case it returns None.
    This can be used to personalize errors.

    If the provided string is a directory, it checks the files inside it.
    if there is only one file inside, it returns said file;
    if there are more files, it tries to filter them with the `filters` keyword(s) to return a single file.
    If this fails, try using more strict filers to return a single file.
    """
    if os.path.isfile(filepath):
        return os.path.abspath(filepath)
    elif os.path.isdir(filepath):
        files = get_list(filepath, filters, abspath=True)
    elif return_anyway:
        return None
    else:
        raise FileNotFoundError('Nothing found at ' + str(filepath))
    # Return a single file
    if len(files) == 1:
        return files[0]
    elif return_anyway:
        return None
    elif len(files) == 0:
        raise FileNotFoundError('The following directory is empty (maybe due to the filters):\n' + filepath)
    else:
        raise FileExistsError(f'More than one file found, please apply a more strict filter. Found:\n{files}')    


def get_list(
        folder:str,
        filters=None,
        abspath:bool=True
    ) -> list:
    """Return the files inside a `folder`, applying optional `filters`.

    The full paths are returned by default; to get only the base names, set `abspath = False`.
    """
    if os.path.isfile(folder):
        folder = os.path.dirname(folder)
    if not os.path.isdir(folder):
        raise FileNotFoundError('Directory not found: ' + folder)
    folder = os.path.abspath(folder)
    files = os.listdir(folder)
    # Apply filters or not
    if filters is not None:
        target_files = []
        if not isinstance(filters, list):
            filters = [str(filters)]
        for filter_i in filters:
            filter_i = os.path.basename(filter_i)
            for f in files:
                if filter_i in f:
                    target_files.append(f)
        files = target_files
    if abspath:
        filepaths = []
        for f in files:
            filepaths.append(os.path.join(folder, f))
        files = filepaths
    return files


def get_dir(folder=None) -> str:
    """Returns the full path of `folder` or the parent folder if it's a file. If none is provided, the current working directory is returned."""
    if folder == None:
        path = os.getcwd()
    elif os.path.isdir(folder):
        path = os.path.realpath(folder)
    elif not os.path.isdir(folder):
        if os.path.isfile:
            path = os.path.dirname(folder)
            path = os.path.realpath(path)
        else:
            raise FileNotFoundError(f'Missing folder at {folder}')
    return path


def copy(
        old:str,
        new:str
    ) -> None:
    """Copies `old` file to `new` file"""
    # Yes, I know, why use ATON for this right? copy() and move() functions are here just for consistancy.
    file = shutil.copy(old, new)
    return None


def move(
        old:str,
        new:str
    ) -> None:
    """Moves `old` file to `new` file."""
    file = shutil.move(old, new)
    return None


def remove(filepath:str) -> None:
    """Removes the given file or folder at `filepath`.

    > WARNING: Removing stuff is always dangerous, be careful!
    """
    if filepath is None:
        return None  # It did not exist in the first place
    elif os.path.isfile(filepath):
        os.remove(filepath)
    elif os.path.isdir(filepath):
        shutil.rmtree(filepath)
    else:
        return None  # It did not exist in the first place
    return None


def rename_on_folder(
        old:str,
        new:str,
        folder=None,
    ) -> None:
    """Batch renames files in the given `folder`.

    Replaces the `old` string by `new` string.
    If no folder is provided, the current working directory is used.
    """
    if folder is None:
        files = os.listdir('.')
    elif os.path.isdir(folder):
        file_list = os.listdir(folder)
        files = []
        for file in file_list:
            file_path = os.path.join(folder, file)
            files.append(file_path)
    elif os.path.isdir(os.path.join(os.getcwd(), folder)):
        folder_path = os.path.join(os.getcwd(), folder)
        file_list = os.listdir(folder_path)
        files = []
        for file in file_list:
            file_path = os.path.join(folder_path, file)
            files.append(file_path)
    else:
        raise FileNotFoundError('Missing folder at ' + folder + ' or in the CWD ' + os.getcwd())
    for f in files:
        if old in f:
            os.rename(f, f.replace(old, new))
    return None


def rename_on_folders(
        old:str,
        new:str,
        folder=None,
    ) -> None:
    """Renames the files inside the subfolders in the parent `folder`.
    
    Renames from an `old` string to the `new` string.
    If no `folder` is provided, the current working directory is used.
    """
    if folder is None:
        things = os.listdir('.')
    elif os.path.isdir(folder):
        things = os.listdir(folder)
    elif os.path.isdir(os.path.join(os.getcwd(), folder)):
        things = os.listdir(os.path.join(os.getcwd(), folder))
    else:
        raise FileNotFoundError('Missing folder at ' + folder + ' or in the CWD ' + os.getcwd())
    for d in things:
        if os.path.isdir(d):
            for f in os.listdir(d):
                if old in f:
                    old_file = os.path.join(d, f)
                    new_file = os.path.join(d, f.replace(old, new))
                    os.rename(old_file, new_file)
    return None


def copy_to_folders(
        folder=None,
        extension:str=None,
        strings_to_delete:list=[],
    ) -> None:
    """Copies the files from the parent `folder` with the given `extension` to individual subfolders.

    The subfolders are named as the original files,
    removing the strings from the `strings_to_delete` list.
    If no `folder` is provided, it runs in the current working directory.
    """
    if folder is None:
        folder = os.getcwd()
    old_files = get_list(folder, extension)
    if old_files is None:
        raise ValueError('No ' + extension + ' files found in path!')
    for old_file in old_files:
        new_file = old_file
        for string in strings_to_delete:
            new_file = new_file.replace(string, '')
        path = new_file.replace(extension, '')
        os.makedirs(path, exist_ok=True)
        new_file_path = os.path.join(path, new_file)
        copy(old_file, new_file_path)
    return None


def save(object, filename:str=None):
    """Save a Python object in the current working directory as a binary `*.aton` file."""
    filename = 'data' if filename is None else filename
    if not filename.endswith('.aton'):
        filename += '.aton'
    file = os.path.join(os.getcwd(), filename)
    with gzip.open(file, 'wb') as f:
        pickle.dump(object, f)
    print(f"Data saved and compressed to {file}")


def load(filepath:str='data.aton'):
    """Load a Python object from a binary `*.aton` file.
    
    Use only if you trust the person who sent you the file!
    """
    file_path = get(filepath)
    with gzip.open(file_path, 'rb') as f:
        data = pickle.load(f)
    return data




"""
# Description

This module is used to solve the hamiltonian eigenvalues and eigenvectors for a given quantum system.
Sparse matrices are used to achieve optimal performance.


# Index

| | |
| --- | --- |
| `laplacian_matrix()`      | Calculate the second derivative matrix for a given grid |
| `hamiltonian_matrix()`    | Calculate the hamiltonian matrix of the system |
| `potential()`             | Solve the potential values of the system |
| `schrodinger()`           | Solve the Schrödiger equation for the system |
| `energies()`              | Solve the system(s) for the `QSys` or `QExp` object |

---
"""


from .classes import *
from .potential import solve as solve_potential
from copy import deepcopy
import time
import numpy as np
from scipy import sparse
import aton


def laplacian_matrix(grid):
    """Calculates the Laplacian (second derivative) matrix for a given `grid`."""
    x = grid
    diagonals = [-2*np.ones(len(x)), np.ones(len(x)), np.ones(len(x))]
    laplacian_matrix = sparse.spdiags(diagonals, [0, -1, 1], format='lil')
    # Periodic boundary conditions
    laplacian_matrix[0, -1] = 1
    laplacian_matrix[-1, 0] = 1
    dx = x[1] - x[0]
    laplacian_matrix /= dx**2
    return laplacian_matrix


def hamiltonian_matrix(system:QSys):
    """Calculates the Hamiltonian matrix for a given `aton.qrotor.classes.QSys` object."""
    V = system.potential_values.tolist()
    potential = sparse.diags(V, format='lil')
    B = system.B
    x = system.grid
    H = -B * laplacian_matrix(x) + potential
    return H


def potential(system:QSys) -> QSys:
    """Solves the potential_values of the system.

    It uses the potential name, by calling `aton.qrotor.potential.solve`.
    Then it applies extra operations, such as removing the potential offset
    if `aton.qrotor.classes.QSys.correct_potential_offset = True`.
    """
    V = solve_potential(system)
    if system.correct_potential_offset is True:
        offset = min(V)
        V = V - offset
        system.corrected_potential_offset = offset
    system.potential_values = V
    return system


def schrodinger(system:QSys) -> QSys:
    """Solves the Schrödinger equation for a given `aton.qrotor.classes.QSys` object.
    
    Uses ARPACK in shift-inverse mode to solve the hamiltonian sparse matrix.
    """
    time_start = time.time()
    V = system.potential_values
    H = hamiltonian_matrix(system)
    print(f'Solving Hamiltonian matrix of size {system.gridsize}...')
    # Solve eigenvalues with ARPACK in shift-inverse mode, with a sparse matrix
    eigenvalues, eigenvectors = sparse.linalg.eigsh(H, system.E_levels, which='LM', sigma=0, maxiter=10000)
    if any(eigenvalues) is None:
        print('WARNING:  Not all eigenvalues were found.\n')
    else: print('Done.\n')
    system.runtime = time.time() - time_start
    system.eigenvalues = eigenvalues
    system.potential_max = max(V)
    system.potential_min = min(V)
    system.energy_barrier = max(V) - min(eigenvalues)
    system.first_transition = eigenvalues[1] - eigenvalues[0]
    if system.save_eigenvectors == True:
        system.eigenvectors = eigenvectors
    system.eigenvalues_B = eigenvalues / system.B
    system.potential_max_B = system.potential_max / system.B
    return system


def energies(var, filename:str=None) -> QExp:
    """Solves the Schrödinger equation for a given `aton.qrotor.classes.QSys` or `aton.qrotor.classes.QExp` object."""
    if isinstance(var, QSys):
        data = QExp()
        data.systems = [deepcopy(var)]
    elif isinstance(var, QExp):
        data = deepcopy(var)
    else:
        raise ValueError('Input must be a QSys or QExp object.')
    for system in data.systems:
        system = potential(system)
        system = schrodinger(system)
    if filename:
        aton.st.file.save(data, filename)
    return data




"""
# Description

This submodule contains tools to rotate molecular structures.
Works with Quantum ESPRESSO input files.


# Index

| | |
| --- | --- |
| `qe()`            | Rotate specific atoms from a Quantum ESPRESSO input file |
| `rotate_coords()` | Rotate a specific list of coordinates |

---
"""


import numpy as np
import os
import shutil
from scipy.spatial.transform import Rotation
from .constants import *
import aton.interface as interface
import aton.txt.extract as extract
import aton.txt.edit as edit


def qe(
        filepath:str,
        positions:list,
        angle:float,
        repeat:bool=False,
        precision:int=3,
        show_axis:bool=False,
    ) -> list:
    """Rotates atoms from a Quantum ESPRESSO input file.

    Takes a `filepath` with a molecular structure, and three or more atomic `positions` (list).
    These input positions can be approximate, and are used to identify the target atoms.
    The decimal precision in the search for these positions is controlled by `precision`.

    It rotates the atoms by the geometrical center of the first 3 atoms by a specific `angle`.
    Additionally, if `repeat = True` it repeats the same rotation over the whole circunference.
    Finally, it writes the rotated structure(s) to a new structural file(s).
    Returns a list with the output filename(s).

    To debug, `show_axis = True` adds two additional helium atoms as the rotation vector.
    """
    print('Rotating Quantum ESPRESSO input structure with QRotor...')
    if len(positions) < 3:
        raise ValueError("At least three positions are required to define the rotation axis.")
    lines = []
    full_positions = []
    for position in positions:
        line = interface.qe.get_atom(filepath, position, precision)
        lines.append(line)
        pos = extract.coords(line)
        if len(pos) > 3:  # Keep only the first three coordinates
            pos = pos[:3]
        # Convert to cartesian
        pos_cartesian = interface.qe.to_cartesian(filepath, pos)
        full_positions.append(pos_cartesian)
        print(f'Found atom: "{line}"')
    # Set the angles to rotate
    if not repeat:
        angles = [angle]
    else:
        angles = range(0, 360, angle)
    # Rotate and save the structure
    outputs = []
    path = os.path.dirname(filepath)
    basename = os.path.basename(filepath)
    name, ext = os.path.splitext(basename)
    print('Rotating the structure...')
    for angle in angles:
        output_name = name + f'_{angle}' + ext
        output = os.path.join(path, output_name)
        rotated_positions_cartesian = rotate_coords(full_positions, angle, show_axis)
        rotated_positions = []
        for coord in rotated_positions_cartesian:
            pos = interface.qe.from_cartesian(filepath, coord)
            rotated_positions.append(pos)
        _save_qe(filepath, output, lines, rotated_positions)
        outputs.append(output)
        print(output)
    return outputs


def rotate_coords(
        positions:list,
        angle:float,
        show_axis:bool=False,
    ) -> list:
    """Rotates geometrical coordinates.

    Takes a list of atomic `positions` in cartesian coordinates, as
    `[[x1,y1,z1], [x2,y2,z2], [x3,y3,z3]], [etc]`.
    Then rotates said coordinates by a given `angle` (degrees),
    taking the perpendicular axis that passes through the
    geometrical center of the first three points as the axis of rotation.
    Any additional coordinates are rotated with the same rotation matrix.
    Returns a list with the updated positions.

    If `show_axis = True` it returns two additional coordinates at the end of the list,
    with the centroid and the rotation vector.
    """
    if len(positions) < 3:
        raise ValueError("At least three coordinates are required to define the rotation axis.")
    if not isinstance(positions[0], list):
        raise ValueError(f"Atomic positions must have the form: [[x1,y1,z1], [x2,y2,z2], [x3,y3,z3], etc]. Yours were:\n{positions}")
    positions = np.array(positions)
    #print(f'POSITIONS: {positions}')  # DEBUG
    # Define the geometrical center of the first three points
    center = np.mean(positions[:3], axis=0)
    # Ensure the axis passes through the geometrical center
    centered_positions = positions - center
    # Define the perpendicular axis (normal to the plane formed by the first three points)
    v1 = centered_positions[0] - centered_positions[1]
    v2 = centered_positions[0] - centered_positions[2]
    # Cross product
    axis = np.cross(v2, v1)
    axis_length = np.linalg.norm(axis)
    axis = axis / axis_length
    # Create the rotation object using scipy
    rotation = Rotation.from_rotvec(np.radians(angle) * axis)
    # Rotate all coordinates around the geometrical center
    rotated_centered_positions = rotation.apply(centered_positions)
    rotated_positions = (rotated_centered_positions + center).tolist()
    #print(f'ROTATED_POSITIONS: {rotated_positions}')  # DEBUG
    if show_axis:
        rotated_positions.append(center.tolist())
        rotated_positions.append((center + axis).tolist())
    return rotated_positions


def _save_qe(
        filename,
        output:str,
        lines:list,
        positions:list
    ) -> str:
    """Copies `filename` to `output`, updating the old `lines` with the new `positions`."""
    shutil.copy(filename, output)
    for i, line in enumerate(lines):
        strings = line.split()
        atom = strings[0]
        new_line = f"  {atom}   {positions[i][0]:.15f}   {positions[i][1]:.15f}   {positions[i][2]:.15f}"
        #print(f'OLD LINE: {line}')  # DEBUG
        #print(f'NEW_LINE: {new_line}')  # DEBUG
        edit.replace_line(output, line, new_line, raise_errors=True)
    if len(lines) == len(positions):
        return output
    elif len(lines) + 2 != len(positions):
        raise ValueError(f"What?!  len(lines)={len(lines)} and len(positions)={len(positions)}")
    # This is only for the show_axis=True case
    additional_positions = positions[-2:]
    for pos in additional_positions:
        pos.insert(0, 'He')
        interface.qe.add_atom(output, pos)
    return output

